from typing import TYPE_CHECKING, Optional, Sequence, Union

from skt.vault_utils import get_secrets

if TYPE_CHECKING:
    from pyspark import SparkConf, SparkContext
    from pyspark.sql import SparkSession


def get_hms():
    from hmsclient import hmsclient

    s = get_secrets(path="ye/hivemetastore")
    host = s["ip"]
    port = s["port"]
    client = hmsclient.HMSClient(host=host, port=port)
    client.open()
    return client


def get_hive_conn():
    from pyhive import hive

    hiveserver2 = get_secrets(path="ye/hiveserver2")
    host = hiveserver2["ip"]
    port = hiveserver2["port"]
    user = hiveserver2["user"]
    conn = hive.connect(host, port=port, username=user)
    return conn


def get_hdfs_conn():
    try:
        from pyarrow import hdfs

        conn = hdfs.connect(user="airflow")
    except ImportError:
        import os
        import subprocess

        from pyarrow import fs

        print("Using pyarrow.fs.HadoopFileSystem, Cause pyarrow.hdfs is deprecated.")

        hadoop_bin = os.path.join(os.environ["HADOOP_HOME"], "bin", "hadoop")
        classpath = subprocess.check_output((hadoop_bin, "classpath", "--glob"))

        os.environ["CLASSPATH"] = classpath.decode("utf-8")

        conn = fs.HadoopFileSystem("default", user="airflow")

    return conn


def get_sqlalchemy_engine():
    from sqlalchemy import create_engine

    hiveserver2 = get_secrets(path="ye/hiveserver2")
    host = hiveserver2["ip"]
    port = hiveserver2["port"]
    user = hiveserver2["user"]
    return create_engine(f"hive://{user}@{host}:{port}/tmp")


def get_pkl_from_hdfs(pkl_path):
    import pickle

    conn = get_hdfs_conn()
    byte_object = conn.cat(f"{pkl_path}")
    pkl_object = pickle.loads(byte_object)
    return pkl_object


def is_in_notebook():
    try:
        from IPython import get_ipython

        if get_ipython().__class__.__name__ == "ZMQInteractiveShell":
            return True
        else:
            return False
    except (NameError, ModuleNotFoundError):
        return False


def _spark_context_to_repr_html(sc: "SparkContext") -> str:
    import os

    if os.environ.get("JUPYTERHUB_USER"):
        ui_web_url = sc.uiWebUrl

    else:
        history_server_url = os.environ.get("SPARK_HISTORY_SERVER_URL", "http://spark.yks.sktai.io")
        ui_web_url = f"{history_server_url}/history/{sc.applicationId}"

    return f"""
        <div>
            <p><b>SparkContext</b></p>

            <p><a href="{ui_web_url}">Spark UI</a></p>

            <dl>
              <dt>Version</dt>
                <dd><code>v{sc.version}</code></dd>
              <dt>Master</dt>
                <dd><code>{sc.master}</code></dd>
              <dt>AppName</dt>
                <dd><code>{sc.appName}</code></dd>
            </dl>
        </div>
        """


def get_spark(
    scale: int = 0,
    queue: Optional[str] = None,
    jars: Optional[Union[Sequence[str], str]] = None,
    conf: Optional["SparkConf"] = None,
) -> "SparkSession":
    import json
    import os
    import uuid
    import warnings

    from pyspark.conf import SparkConf
    from pyspark.sql import SparkSession

    if queue:
        warnings.warn("Queue is unsupported not implemented yet")

    default = {
        "spark.app.name": f"skt-{os.environ.get('USER', 'default')}-{str(uuid.uuid4())}",
        "spark.driver.cores": "1",
        "spark.executor.cores": "1",
    }

    if jars:
        if isinstance(jars, str):
            default["spark.jars"] = jars
        else:
            default["spark.jars"] = ",".join(jars)

    if not scale:
        default["spark.driver.memory"] = "8g"
        default["spark.driver.maxResultSize"] = "6g"
        default["spark.executor.memory"] = "8g"
        default["spark.executor.instances"] = "8"
    elif 1 <= scale <= 4:
        default["spark.driver.memory"] = f"{scale * 8}g"
        default["spark.driver.maxResultSize"] = f"{scale * 4}g"
        default["spark.executor.memory"] = f"{scale * 3}g"
        default["spark.executor.instances"] = str(scale * 8)
    elif scale < 9:
        default["spark.driver.memory"] = "8g"
        default["spark.driver.maxResultSize"] = "8g"
        default["spark.executor.memory"] = f"{min(2**scale, 200)}g"
        default["spark.executor.instances"] = "32"
    else:
        raise ValueError("Invalid scale")

    if not conf:
        conf = SparkConf()

    for key, value in default.items():
        if not conf.get(key):
            conf.set(key, value)

    task_group = [
        {
            "name": "executor",
            "minMember": int(conf.get("spark.executor.instances")),
            "minResource": {
                "cpu": conf.get("spark.executor.cores"),
                "memory": f'{int(conf.get("spark.executor.memory").replace("g","")) * 1.1}Gi',  # JVM Overhead
            },
            "nodeSelector": {
                "role/spark": None,
                "zone": "cpu.pool",
            },
            "tolerations": [],
            "affinity": {},
            "topologySpreadConstraints": [],
        }
    ]

    conf = (
        conf.set("spark.kubernetes.executor.annotation.yunikorn.apache.org/task-group-name", "executor")
        .set("spark.kubernetes.executor.annotation.yunikorn.apache.org/task-groups", json.dumps(task_group))
        .set(
            "spark.kubernetes.executor.annotation.yunikorn.apache.org/schedulingPolicyParameters",
            "placeholderTimeoutInSeconds=60 gangSchedulingStyle=Hard",
        )
    )

    spark = SparkSession.builder.config(conf=conf).getOrCreate()

    if is_in_notebook():
        from IPython.display import HTML, display

        display(HTML(_spark_context_to_repr_html(spark.sparkContext)))

    return spark


def hive_execute(query):
    conn = get_hive_conn()
    c = conn.cursor()
    c.execute(query)
    c.close()
    conn.close()


def hive_get_result(query):
    conn = get_hive_conn()
    c = conn.cursor()
    c.execute(query)
    result = c.fetchall()
    c.close()
    conn.close()
    return result


def hive_to_pandas(query, scale=0):
    if scale == 1:
        import pandas

        conn = get_hive_conn()
        df = pandas.read_sql(query, conn)
        df.info()
        conn.close()
        return df

    import uuid

    tmp_id = str(uuid.uuid4()).replace("-", "_")
    ctas = f"CREATE TABLE dumbo.{tmp_id} stored as parquet as {query}"
    conn = get_hive_conn()
    c = conn.cursor()
    c.execute("set parquet.column.index.access=false")
    c.execute(ctas)
    df = parquet_to_pandas(f"/warehouse/tablespace/managed/hive/dumbo.db/{tmp_id}")
    c.execute(f"DROP TABLE dumbo.{tmp_id}")
    c.close()
    conn.close()
    return df


def parquet_to_pandas(hdfs_path):
    import os

    from pyarrow import fs, parquet

    from skt.ye import get_hdfs_conn

    # Load hadoop environment
    get_hdfs_conn().close()

    os.environ["ARROW_LIBHDFS_DIR"] = "/usr/hdp/3.0.1.0-187/usr/lib"
    hdfs = fs.HadoopFileSystem("ye.sktai.io", 8020, user="airflow")
    if hdfs_path.startswith("hdfs://yellowelephant/"):
        hdfs_path = hdfs_path[21:]
    df = parquet.read_table(hdfs_path, filesystem=hdfs).to_pandas()
    df.info()
    return df


def pandas_to_parquet(pandas_df, hdfs_path, spark):
    df = spark.createDataFrame(pandas_df)
    df.write.mode("overwrite").parquet(hdfs_path)


def slack_send(
    text="This is default text",
    username="SKT",
    channel="#leavemealone",
    icon_emoji=":large_blue_circle:",
    blocks=None,
    dataframe=False,
    adot=False,
):
    import requests

    from skt.vault_utils import get_secrets

    if dataframe:
        from tabulate import tabulate

        text = "```" + tabulate(text, tablefmt="simple", headers="keys") + "```"

    token = (
        get_secrets("airflow_k8s/adot_slack/slack_alarmbot_token")["token"]
        if adot
        else get_secrets("slack")["bot_token"]["airflow"]
    )
    proxy = get_secrets("proxy")["proxy"]
    proxies = {
        "http": proxy,
        "https": proxy,
    }
    headers = {
        "Content-Type": "application/json;charset=utf-8",
        "Authorization": f"Bearer {token}",
    }
    json_body = {
        "username": username,
        "channel": channel,
        "text": text,
        "blocks": blocks,
        "icon_emoji": icon_emoji,
    }
    r = requests.post(
        "https://www.slack.com/api/chat.postMessage",
        proxies=proxies,
        headers=headers,
        json=json_body,
    )
    r.raise_for_status()
    if not r.json()["ok"]:
        raise Exception(r.json())


def send_email(subject, text, send_from, send_to, attachment=None):
    """
    :param str attachment: Attachment to send as .txt file with email
    """
    import smtplib
    from email.mime.application import MIMEApplication
    from email.mime.multipart import MIMEMultipart
    from email.mime.text import MIMEText
    from email.utils import formatdate

    from skt.vault_utils import get_secrets

    c = get_secrets(path="mail")
    host, port = c["smtp_host"], c["smtp_port"]
    msg = MIMEMultipart()
    msg["From"] = send_from
    msg["To"] = send_to
    msg["Date"] = formatdate(localtime=True)
    msg["Subject"] = subject
    msg.attach(MIMEText(text))

    if attachment:
        part = MIMEApplication(attachment, NAME=subject)
        part.add_header("Content-Disposition", f"attachment; filename={subject}.txt")
        msg.attach(part)

    with smtplib.SMTP(host, port) as smtp:
        if "sktelecom.com" in host:
            smtp.starttls()
        return smtp.sendmail(send_from, send_to.split(","), msg.as_string())


def get_github_util():
    from skt.github_utils import GithubUtil

    github_token = get_secrets("github/sktaiflow")["token"]
    proxy = get_secrets("proxy")["proxy"]
    proxies = {
        "http": proxy,
        "https": proxy,
    }
    g = GithubUtil(github_token, proxies)
    return g


def _write_to_parquet_via_spark(pandas_df, hdfs_path):
    spark = get_spark()
    spark_df = spark.createDataFrame(pandas_df)
    spark_df.write.mode("overwrite").parquet(hdfs_path)


def _write_to_parquet(pandas_df, hdfs_path):
    import pyarrow as pa
    import pyarrow.parquet as pq

    # Read Parquet INT64 timestamp issue:
    # https://issues.apache.org/jira/browse/HIVE-21215
    if "datetime64[ns]" in pandas_df.dtypes.tolist():
        _write_to_parquet_via_spark(pandas_df, hdfs_path)
        return

    pa_table = pa.Table.from_pandas(pandas_df)
    hdfs_conn = get_hdfs_conn()
    try:
        pq.write_to_dataset(pa_table, root_path=hdfs_path, filesystem=hdfs_conn)
    finally:
        hdfs_conn.close()


def _write_df(pandas_df, schema_name, table_name, hdfs_path, engine, cursor, tmp_table_name):
    import sqlalchemy.exc

    cursor.execute(f"drop table if exists {schema_name}.{tmp_table_name}")
    try:
        pandas_df.to_sql(tmp_table_name, engine, schema=schema_name, if_exists="replace", index=False)
    except sqlalchemy.exc.ProgrammingError:
        # Hive bulk insert issue:
        # https://github.com/dropbox/PyHive/issues/343
        pass

    cursor.execute(f"drop table if exists {schema_name}.{table_name}")
    if hdfs_path is None:
        cursor.execute(
            f"""create table {schema_name}.{table_name}
                           like {schema_name}.{tmp_table_name}
                           stored as parquet"""
        )
        cursor.execute(f"show create table {schema_name}.{table_name}")
        result = cursor.fetchall()
        managed_hdfs_path = list(filter(lambda row: row[0].strip().find("hdfs://") == 1, result))[0][0].strip()[1:-1]
        _write_to_parquet(pandas_df, managed_hdfs_path)
    else:
        cursor.execute(
            f"""create external table {schema_name}.{table_name}
                           like {schema_name}.{tmp_table_name}
                           stored as parquet
                           location '{hdfs_path}'"""
        )


def write_df_to_hive(pandas_df, schema_name, table_name, hdfs_path=None):
    """
    Exports a Panadas dataframe into a table in Hive.

    Example:
    write_df_to_hive(pandas_df1, "my_schema", "my_table1")
    write_df_to_hive(pandas_df2, "my_schema", "my_table2")
    write_df_to_hive(pandas_df1, "my_schema", "my_table3",
            hdfs_path="hdfs://.../my_schema.db/my_table1")

    Parameters
    ----------
    pandas_df : an ojbect of Pandas Dataframe
    schema_name : str
        A target schema name of Hive
    table_name : str
        A target table name of Hive
    hdfs_path : str, default None
        A path of Hadoop file system as an optional parameter.
        It will be used to create an external table. If hdfs_path
        is not None, data in the dataframe will not be converted.
        A metadata in the dataframe is just used to create a Hive
        table.
    """
    engine = get_sqlalchemy_engine()
    conn = get_hive_conn()
    cursor = conn.cursor()

    import hashlib

    tmp_table_name = hashlib.sha1(str(f"{schema_name}.{table_name}").encode("utf-8")).hexdigest()

    try:
        _write_df(pandas_df, schema_name, table_name, hdfs_path, engine, cursor, tmp_table_name)
    finally:
        cursor.execute(f"drop table if exists {schema_name}.{tmp_table_name}")
        cursor.close()
        conn.close()



import os
import uuid
from typing import Union

import pandas as pd
from sqlalchemy import text

from skt.gcp import (
    PROJECT_ID,
    bq_table_exists,
    gcs_to_bq,
    get_bigquery_table_partition_info,
)


def get_trino_engine(
    cluster_name: str = None,
    host: str = "gateway-idp-prd.sktai.io",
    port: int = 443,
    connect_args: dict = None,
    user: str = None,
    password: str = None,
    extra_connect_args: dict = None,
):
    from sqlalchemy import create_engine
    from trino.sqlalchemy import URL

    nb_user = os.environ.get("NB_USER", "skt")

    # jovyan: NES로 실행 시 Jupyter 이미지일 경우 NB_USER가 jovyan으로 설정된다.
    # skt: 임의의 장소에서 패키지를 실행했을 경우 NB_USER 값이 없을 수 있는데, 이럴 경우 skt로 설정한다.

    if cluster_name is None:
        cluster_name = "aidp-cluster" if nb_user in ["jovyan", "skt"] else "aidp-interactive"

    client_tag = "NES" if nb_user in ["jovyan", "skt"] else "jupyter"

    extra_connect_args = extra_connect_args or {}
    connect_args = connect_args or {
        "extra_credential": [("cluster_name", cluster_name)],
        "http_scheme": "https",
        "client_tags": [client_tag],
        **extra_connect_args,
    }

    return create_engine(
        URL(host=host, port=port, user=user or nb_user, password=password),
        connect_args=connect_args,
    )


def init_trino(
    cluster_name: str = None,
    host: str = "gateway-idp-prd.sktai.io",
    port: int = 443,
    connect_args: dict = None,
    user: str = None,
    password: str = None,
    extra_connect_args: dict = None,
):
    engine = get_trino_engine(cluster_name, host, port, connect_args, user, password, extra_connect_args)

    try:
        from IPython import get_ipython

        ipython = get_ipython()
        if ipython:
            ipython.run_line_magic("load_ext", "sql", 1)
            ipython.run_line_magic("alias_magic", "trino sql", 1)
            ipython.run_line_magic("sql", "engine", 1)
    except ImportError:
        print("IPython not found, skipping magic commands")

    return engine


def _overwrite_trino_query_result_to_bq_table_partition(
    trino_query, temp_table_name, bigquery_dataset, bigquery_table, partition_info, project_id=PROJECT_ID
):
    engine = get_trino_engine()

    with engine.connect() as conn:
        insert_phrase = f"""create table "ye"."gcs_temp_seoul_1d"."{temp_table_name}" with (format = 'PARQUET') as\n"""
        insert_query = insert_phrase + trino_query
        gcs_temp_table_ctas_result = conn.execute(text(insert_query))
        print(gcs_temp_table_ctas_result.fetchall()[0][0], "rows overwrite")

        temp_table_range_distinct = (
            f"""select distinct({partition_info['field']}) from "ye"."gcs_temp_seoul_1d"."{temp_table_name}"\n"""
        )
        temp_table_range_distinct_proxy = conn.execute(text(temp_table_range_distinct))
        temp_table_range_distinct_values = temp_table_range_distinct_proxy.fetchall()
        temp_table_partition_values = [row[0] for row in temp_table_range_distinct_values]
        for delete_partition in temp_table_partition_values:
            if partition_info["partitioning"] == "time_partitioning":
                delete_partition_str = delete_partition.strftime("%Y-%m-%d")
                delete_query = f"""delete from "bigquery"."{bigquery_dataset}"."{bigquery_table}" where {partition_info['field']}=DATE('{delete_partition_str}')"""
                conn.execute(text(delete_query))
                print(
                    f"partition_column: {partition_info['field']}", f"overwrite partition value: {delete_partition_str}"
                )
            elif partition_info["partitioning"] == "range_partitioning":
                delete_query = f"""delete from "bigquery"."{bigquery_dataset}"."{bigquery_table}" where {partition_info['field']}={delete_partition}"""
                conn.execute(text(delete_query))
                print(f"partition_column: {partition_info['field']}", f"overwrite partition value: {delete_partition}")

        conn.commit()
        gcs_to_bq(temp_table_name, "temp-seoul-1d", bigquery_dataset, bigquery_table, overwrite=False)


def load_trino_query_result_to_bq_table(
    trino_query, bigquery_dataset, bigquery_table, overwrite=False, project_id=PROJECT_ID
):
    is_bq_table_exist = bq_table_exists(f"{bigquery_dataset}.{bigquery_table}", project_id=project_id)
    temp_table_name = "temp_" + str(uuid.uuid4()).replace("-", "_")

    if is_bq_table_exist and overwrite:
        partition_info = get_bigquery_table_partition_info(bigquery_dataset, bigquery_table)
        if partition_info:
            _overwrite_trino_query_result_to_bq_table_partition(
                trino_query, temp_table_name, bigquery_dataset, bigquery_table, partition_info
            )
        else:
            load_trino_query_result_to_ye_gcs(trino_query, temp_table_name)
            gcs_to_bq(temp_table_name, "temp-seoul-1d", bigquery_dataset, bigquery_table, overwrite=True)
            return
    else:
        load_trino_query_result_to_ye_gcs(trino_query, temp_table_name)
        gcs_to_bq(temp_table_name, "temp-seoul-1d", bigquery_dataset, bigquery_table, overwrite=False)
        return


def load_trino_query_result_to_ye_gcs(trino_query, ye_table_name, ye_schema_name="gcs_temp_seoul_1d") -> str:
    engine = get_trino_engine()
    tables_list_df = show_tables("ye", ye_schema_name)
    table_list = list(tables_list_df["Table"])
    insert_phrase = None
    if ye_table_name in table_list:
        insert_phrase = f"""insert into "ye"."{ye_schema_name}"."{ye_table_name}"\n"""
    else:
        insert_phrase = f"""create table "ye"."{ye_schema_name}"."{ye_table_name}" with (format = 'PARQUET') as\n"""
    insert_query = insert_phrase + trino_query
    with engine.connect() as conn:
        conn.execute(text(insert_query))
        conn.commit()

    return f'''"ye"."{ye_schema_name}"."{ye_table_name}"'''


def get_trino_query_execution_result(query: str, cluster_name: str = None) -> pd.DataFrame:
    query_text = text(query)
    engine = get_trino_engine(cluster_name=cluster_name)
    with engine.begin() as conn:
        result = conn.execute(query_text)
        columns = result.keys()
        data = result.fetchall()
        df = pd.DataFrame(data, columns=columns)

    return df


def execute_trino_query(query: str, cluster_name: str = None):
    query_text = text(query)
    engine = get_trino_engine(cluster_name=cluster_name)
    with engine.connect() as conn:
        conn.execute(query_text)
        conn.commit()

    return


def show_catalogs(cluster_name: str = None) -> pd.DataFrame:
    query = "show catalogs"
    print(query)
    return get_trino_query_execution_result(query, cluster_name)


def show_schemas(catalog_name: str, cluster_name: str = None) -> pd.DataFrame:
    query = f'''show schemas from "{catalog_name}"'''
    print(query)
    return get_trino_query_execution_result(query, cluster_name)


def show_tables(catalog_name: str, schema_name: str, cluster_name: str = None) -> pd.DataFrame:
    query = f'''show tables from "{catalog_name}"."{schema_name}"'''
    print(query)
    return get_trino_query_execution_result(query, cluster_name)


def show_table(catalog_name: str, schema_name: str, table_name: str, cluster_name: str = None) -> pd.DataFrame:
    query = f'''show tables from "{catalog_name}"."{schema_name}"."{table_name}"'''
    print(query)
    return get_trino_query_execution_result(query, cluster_name)


def create_iceberg_table_from_trino_query_result(
    trino_query: str, catalog_name: str, schema_name: str, table_name: str, partition_column: Union[str, list] = None
):
    partition_column_string = None
    if partition_column is None:
        pass
    elif isinstance(partition_column, list):
        partition_column_string = ", ".join(f"'{str(col)}'" for col in partition_column)
    elif isinstance(partition_column, str):
        partition_column_string = f"'{partition_column}'"
    else:
        raise ValueError("partition_column should be either a string or a list")

    create_table_phrase = f'''create table "{catalog_name}"."{schema_name}"."{table_name}"'''
    if partition_column:
        create_table_phrase += f""" WITH(
        partitioning = ARRAY[{partition_column_string}])"""
    query = create_table_phrase + f""" AS\n{trino_query}"""
    get_trino_query_execution_result(query)


def load_trino_query_result_to_bingo(trino_query: str, bingo_schema_name: str, bingo_table_name: str, overwrite=False):
    tables = show_tables("bingo", bingo_schema_name)
    table_list = list(tables["Table"])
    # 테이블 미 존재시, 생성
    if bingo_table_name not in table_list:
        create_iceberg_table_from_trino_query_result(trino_query, "bingo", bingo_schema_name, bingo_table_name)
        return

    if overwrite:
        delete_query = f'''delete from "bingo"."{bingo_schema_name}"."{bingo_table_name}"'''
        execute_trino_query(delete_query)
        query = f"""insert into "bingo"."{bingo_schema_name}"."{bingo_table_name}"\n{trino_query}"""
        execute_trino_query(query)
    else:
        query = f"""insert into "bingo"."{bingo_schema_name}"."{bingo_table_name}"\n{trino_query}"""
        execute_trino_query(query)


def load_trino_query_result_to_onprem_backup(trino_query: str, table_name: str, overwrite=False):
    tables = show_tables("bingo", "cold_archive")
    table_list = list(tables["Table"])
    # 테이블 미 존재시, 생성
    if table_name not in table_list:
        create_iceberg_table_from_trino_query_result(trino_query, "bingo", "cold_archive", table_name)
        return

    if overwrite:
        delete_query = f'''delete from "bingo"."cold_archive"."{table_name}"'''
        execute_trino_query(delete_query)
        query = f"""insert into "bingo"."cold_archive"."{table_name}"\n{trino_query}"""
        execute_trino_query(query)
    else:
        query = f"""insert into "bingo"."cold_archive"."{table_name}"\n{trino_query}"""
        execute_trino_query(query)


def get_trino_query_result_column_distinct_values(trino_query: str, column_name: str) -> list:
    temp_table_name = "temp_" + str(uuid.uuid4()).replace("-", "_")
    create_iceberg_table_from_trino_query_result(trino_query, "bingo", "temp_1d", temp_table_name)
    query = f'''select distinct "{column_name}" from "bingo"."temp_1d"."{temp_table_name}"'''
    result = get_trino_query_execution_result(query)
    value_list = list(result[column_name])
    return value_list



from enum import Enum
from typing import List, Optional, Union
from uuid import uuid4

from pyspark.sql import DataFrame, SparkSession


class HashType(str, Enum):
    RAW = "raw"
    LAKE_HASH = "lake_hash"
    SHA256 = "sha256"


def skt_hash(
    ss: SparkSession,
    df: DataFrame,
    before: Union[str, HashType],
    after: Union[str, HashType],
    key_column: Optional[Union[str, List[str]]] = None,
    key_type: Optional[Union[str, List[str]]] = None,
) -> DataFrame:
    """

    Transform hashed values between different hash types (raw, lake_hash, sha256).

    Args:
        ss (SparkSession): Spark session object
        df (DataFrame): Input DataFrame containing hashed values
        before (Union[str, HashType]): Current hash type of the values
        after (Union[str, HashType]): Target hash type to convert to
        key_column (Optional[Union[str, List[str]]]): Column name(s) containing hashed values
        key_type (Optional[Union[str, List[str]]]): Type(s) of the key(s) (e.g. 'svc_mgmt_num', 'cust_num')
    Example:
    Returns:
        DataFrame: DataFrame with transformed hash values

    Raises:
        ValueError: If invalid key_type is provided

    Example:
        >>> df = skt_hash(spark, df, 'lake_hash', 'raw', 'service_id', 'svc_mgmt_num')
        # Converts lake_hash values in service_id column to raw values

        >>> df = skt_hash(
        ...     spark,
        ...     df,
        ...     'lake_hash',
        ...     'raw',
        ...     key_column=['svc_mgmt_num', 'cust_num'],
        ...     key_type=['svc_mgmt_num', 'cust_num']
        ... )
        # Converts lake_hash values in service_id and customer_id columns to raw values
    """

    def get_mapping_df(ss: SparkSession, key_type: str) -> DataFrame:
        from pyspark.sql.utils import AnalysisException

        spark = ss
        table_name = f"aidp.{key_type}_mapping"
        try:
            mapping_df = spark.sql(
                f"""
                    SELECT raw, ye_hashed, lake_hashed 
                    FROM {table_name} 
                    WHERE dt = (SELECT MAX(dt) FROM {table_name})
                """
            )
        except AnalysisException as e:
            raise ValueError(f"{key_type}: Invalid key type")
        return mapping_df

    def lake_hash_to_sha256(ss: SparkSession, source_df: DataFrame, key: str, key_type: str) -> DataFrame:
        spark = ss

        source_df.createOrReplaceTempView("source_table")
        mapping_df = get_mapping_df(ss=spark, key_type=key_type)
        mapping_df.createOrReplaceTempView("mapping_table")

        uuid_col = uuid4().hex

        result_df = spark.sql(
            f"""
                SELECT
                    a.*,
                    sha2(b.raw, 256) as {uuid_col}
                FROM
                    source_table a
                    LEFT OUTER JOIN mapping_table b
                    ON a.{key} = b.lake_hashed
            """
        )

        result_df = result_df.drop(key).withColumnRenamed(uuid_col, key)

        spark.catalog.dropTempView("source_table")
        spark.catalog.dropTempView("mapping_table")

        return result_df

    def lake_hash_to_raw(ss: SparkSession, source_df: DataFrame, key: str, key_type: str) -> DataFrame:
        spark = ss

        source_df.createOrReplaceTempView("source_table")
        mapping_df = get_mapping_df(ss=spark, key_type=key_type)
        mapping_df.createOrReplaceTempView("mapping_table")

        uuid_col = uuid4().hex

        result_df = spark.sql(
            f"""
                SELECT
                    a.*,
                    b.raw as {uuid_col}
                FROM
                    source_table a
                    LEFT OUTER JOIN mapping_table b
                    ON a.{key} = b.lake_hashed
            """
        )

        result_df = result_df.drop(key).withColumnRenamed(uuid_col, key)

        spark.catalog.dropTempView("source_table")
        spark.catalog.dropTempView("mapping_table")

        return result_df

    def raw_to_sha256(ss: SparkSession, source_df: DataFrame, key: str, key_type: str) -> DataFrame:
        spark = ss
        source_df.createOrReplaceTempView("source_table")

        uuid_col = uuid4().hex

        result_df = spark.sql(
            f"""
                SELECT
                    a.*,
                    sha2({key}, 256) as {uuid_col}
                FROM
                    source_table a
            """
        )

        result_df = result_df.drop(key).withColumnRenamed(uuid_col, key)

        spark.catalog.dropTempView("source_table")
        return result_df

    def raw_to_lake_hash(ss: SparkSession, source_df: DataFrame, key: str, key_type: str) -> DataFrame:
        spark = ss

        source_df.createOrReplaceTempView("source_table")
        mapping_df = get_mapping_df(ss=spark, key_type=key_type)
        mapping_df.createOrReplaceTempView("mapping_table")

        uuid_col = uuid4().hex

        result_df = spark.sql(
            f"""
                SELECT
                    a.*,
                    b.lake_hashed as {uuid_col}
                FROM
                    source_table a
                    LEFT OUTER JOIN mapping_table b
                    ON a.{key} = b.raw
            """
        )

        result_df = result_df.drop(key).withColumnRenamed(uuid_col, key)

        spark.catalog.dropTempView("source_table")
        spark.catalog.dropTempView("mapping_table")

        return result_df

    def sha256_to_raw(ss: SparkSession, source_df: DataFrame, key: str, key_type: str) -> DataFrame:
        spark = ss

        source_df.createOrReplaceTempView("source_table")
        mapping_df = get_mapping_df(ss=spark, key_type=key_type)
        mapping_df.createOrReplaceTempView("mapping_table")

        uuid_col = uuid4().hex

        result_df = spark.sql(
            f"""
                SELECT
                    a.*,
                    b.raw as {uuid_col}
                FROM
                    source_table a
                    LEFT OUTER JOIN mapping_table b
                    ON a.{key} = b.ye_hashed
            """
        )

        result_df = result_df.drop(key).withColumnRenamed(uuid_col, key)

        spark.catalog.dropTempView("source_table")
        spark.catalog.dropTempView("mapping_table")

        return result_df

    def sha256_to_lake_hash(ss: SparkSession, source_df: DataFrame, key: str, key_type: str) -> DataFrame:
        spark = ss

        source_df.createOrReplaceTempView("source_table")
        mapping_df = get_mapping_df(ss=spark, key_type=key_type)
        mapping_df.createOrReplaceTempView("mapping_table")

        uuid_col = uuid4().hex

        result_df = spark.sql(
            f"""
                SELECT
                    a.*,
                    b.lake_hashed as {uuid_col}
                FROM
                    source_table a
                    LEFT OUTER JOIN mapping_table b
                    ON a.{key} = b.ye_hashed
            """
        )

        result_df = result_df.drop(key).withColumnRenamed(uuid_col, key)

        spark.catalog.dropTempView("source_table")
        spark.catalog.dropTempView("mapping_table")

        return result_df

    if isinstance(before, str):
        before = HashType(before)

    if isinstance(after, str):
        after = HashType(after)

    if key_type is None:
        key_types = ["svc_mgmt_num"]
    elif isinstance(key_type, str):
        key_types = [key_type]
    elif isinstance(key_type, list):
        key_types = key_type

    if key_column is None:
        key_columns = ["svc_mgmt_num"]
    elif isinstance(key_column, str):
        key_columns = [key_column]
    elif isinstance(key_column, list):
        key_columns = key_column

    if len(key_columns) != len(key_types):
        raise ValueError(
            f"Length of key_columns ({len(key_columns)}) and key_types ({len(key_types)}) must be the same"
        )

    conversion_func_name = f"{before.value}_to_{after.value}"

    result_df = df
    for k_col, k_type in zip(key_columns, key_types):
        result_df = eval(conversion_func_name)(ss=ss, source_df=result_df, key=k_col, key_type=k_type)

    return result_df.select(df.columns)


skt_crypto = skt_hash
